/* avr-isp.cc
 * 
 * avr-isp: Programmer software for Atmel AVR family.
 * Copyright (C) 2001 Peter van den Bosch
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * Author:
 * Peter van den Bosch
 * pvdb@dse.nl
 * http://www.dse.nl/~pvdb 
 */

#include <time.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/times.h>
#include <sys/time.h>
#include <sys/io.h>
#include "avr-isp.h"
#include "hexdata.h"

#define DEBUG( args...) printf( "DEBUG: " args)

/* Flags controlling programming mode */
#ifdef	ENDIAN
int FLAG_ENDIAN = 1;
#else
int FLAG_ENDIAN = 0;
#endif
int FLAG_NOVERIFY = 0;

/* Flags to help decide on actions, 
 * not needed in GUI version. */
int FLAG_HELP = 0;
int FLAG_LIST = 0;
int FLAG_VERSION = 0;
int FLAG_READAVR = 0;
int FLAG_WRITEAVR = 0;
int FLAG_WRITEFUSE = 0;
int FLAG_READFUSE = 0;
int FLAG_WRITELOCK = 0;

/* These depend on the AVR that is used, this is for ATmega8 */
char STRFUSE_L1[8][9] =
   { "BODLEVEL", "BODEN", "SUT1", "SUT0", "CKSEL3", "CKSEL2", "CKSEL1",
   "CKSEL0"
};
/* For ATtiny12 */
char STRFUSE_L2[8][9] =
   { "BODLEVEL", "BODEN", "SPIEN", "RSTDISBL", "CKSEL3", "CKSEL2", "CKSEL1",
   "CKSEL0"
};
char STRFUSE_H1[8][9] =
   { "RSTDISBL", "WDTON", "SPIEN", "CKOPT", "EESAVE", "BOOTSZ1",
   "BOOTSZ0",
   "BOOTRST"
};
char STRFUSE_H2[8][9] =
   { "S8515C", "WDTON", "SPIEN", "CKOPT", "EESAVE", "BOOTSZ1", "BOOTSZ0",
   "BOOTRST"
};
char STRFUSE_H3[8][9] =
   { "OCDEN", "JTAGEN", "SPIEN", "CKOPT", "EESAVE", "BOOTSZ1", "BOOTSZ0",
   "BOOTRST"
};

struct ID_CODES id_codes[] = {
   {0x1E9001, "AT90s1200", 1024, 64, 0x00, 0xFF, 0xFF},
   {0x1E9101, "AT90s2313", 2048, 128, 0x80, 0x7F, 0x7F},
   {0x1E9102, "AT90s2323", 2048, 128, 0x00, 0xFF, 0xFF},
   {0x1E9103, "AT90s2343", 2048, 128, 0x00, 0xFF, 0xFF},
   {0x1E9105, "AT90s2333", 2048, 128, 0x00, 0xFF, 0xFF},
   {0x1E9203, "AT90s4433", 4096, 256, 0x00, 0xFF, 0xFF},
   {0x1E9201, "AT90s4414", 4096, 256, 0x80, 0x7F, 0x7F},
   {0x1E9301, "AT90s8515", 8192, 512, 0x80, 0x7F, 0x7F},
   {0x1E9307, "ATmega8", 8192, 512, 0xFF, 0xFF, 0xFF, 32, STRFUSE_L1,
    STRFUSE_H1},
   {0x1E9306, "ATmega8515", 8192, 512, 0xFF, 0xFF, 0xFF, 32, STRFUSE_L1,
    STRFUSE_H2},
   {0x1E9403, "ATmega16", 16384, 512, 0xFF, 0xFF, 0xFF, 64, STRFUSE_L1,
    STRFUSE_H3},
   {0x1E, "ATmega32", 32768, 1024, 0xFF, 0xFF, 0xFF, 64, STRFUSE_L1,
    STRFUSE_H3},
   {0x1E, "ATmega64", 65536, 2048, 0xFF, 0xFF, 0xFF, 128, STRFUSE_L1,
    STRFUSE_H3},
   {0x1E9005, "ATtiny12", 1024, 64, 0x00, 0xFF, 0xFF, 0, STRFUSE_L2},
   {}
};
struct ID_CODES *this_device = NULL;

/** avr_program
 *
 *  Program input from file pointers to flash and EEPROM
 *  Return: 0 - succesfull
 *          -1 - error programming FLASH
 *          -2 - error programming EEPROM
 */
int
avr_program (FILE * fpflash, FILE * fpeeprom, void *(*progress) (int, double))
{
   hexdata *hexfile;
   unsigned char data[128 * 2];
   int index = 0;
   long filesize, writtenbytes = 0L;
   int page, pagesize, nrBytes;

   if (avr_init ())
   {
      avr_run ();
      return -1;
   }

   avr_erase ();

   /* Writing to AVR-FLASH is reqeuested. */
   if (fpflash != NULL)
   {
      hexfile = new hexdata (this_device->FLASHsize);
      if (NULL == hexfile)
      {
         avr_run ();
         return -1;
      }

      /* Read the entire file */
      if ((filesize = hexfile->loadihex (fpflash)) <= 0)
      {
         avr_run ();
         delete hexfile;
         if( filesize == 0)
         {
            return 0;
         } else {
            return -1;
         }
      }

      /* Reset progressbar */
      progress (FLASH, 0.0);

      if (!this_device->pageprog)
      {
         pagesize = 16;
      }
      else
      {
         pagesize = this_device->pageprog * 2;
      }
      /* Check every byte */
      for (page = 0; page < this_device->FLASHsize / pagesize; page++)
      {
         nrBytes = 0;
         for (index = 0; index < pagesize; index++)
         {
            if (hexfile->isused (page * pagesize + index))
            {
               nrBytes++;
               data[index] = hexfile->getdata (page * pagesize + index);
            }
            else
            {
               data[index] = 0xFF;
            }
         }
         if (nrBytes)
         {
            if (download (pagesize, data, FLASH, page * pagesize))
            {
               avr_run ();
               delete hexfile;
               return -1;
            }
            writtenbytes += nrBytes;
            progress (FLASH, (double) writtenbytes / (double) filesize);
         }
      }
      fprintf (stdout, "\n");
      delete hexfile;
   }

   /* Writing to AVR-EEPROM is reqeuested. */
   if (fpeeprom != NULL)
   {
      hexfile = new hexdata (this_device->EEPROMsize);
      if (NULL == hexfile)
      {
         avr_run ();
         return -2;
      }

      /* Read the entire file */
      if ((filesize = hexfile->loadihex (fpeeprom)) < 0)
      {
         delete hexfile;
         avr_run ();
         if( filesize == 0)
         {
            return 0;
         } else {
            return -2;
         }
      }
      
      /* Reset progressbar */
      writtenbytes = 0;
      progress (EEPROM, 0.0);

      pagesize = 16;

      /* Check every byte */
      for (page = 0; page < this_device->EEPROMsize / pagesize; page++)
      {
         nrBytes = 0;
         for (index = 0; index < pagesize; index++)
         {
            if (hexfile->isused (page * pagesize + index))
            {
               nrBytes++;
               data[index] = hexfile->getdata (page * pagesize + index);
            }
            else
            {
               data[index] = 0xFF;
            }
         }
         if (nrBytes)
         {
            if (download (pagesize, data, EEPROM, page * pagesize))
            {
               delete hexfile;
               avr_run ();
               return -2;
            }
            writtenbytes += nrBytes;
            progress (EEPROM, (double) writtenbytes / (double) filesize);
         }
      }
      fprintf (stdout, "\n");
      delete hexfile;
   }

   avr_run ();
   return 0;
}

/** avr_openport( void)
 *
 *  Initialize parallel port for usage
 *  Returns: 0 on succes
 *          -1 on failure
 */
int
avr_openport (void)
{
   /* Set permissions of the IO ports so that we can read & write from and to 
    * it. */
   if (ioperm (LPT_BASE, 3, 0x666) == -1)
   {
      fprintf (stderr, "ERROR: Unable to set permissions on parallel port.\n");
      return -1;
   }

   return 0;
}

/** avr_readfuses( unsigned char *buf)
 * 
 *  Read fuse bytes from AVR
 */
int
avr_readfuses (unsigned char *buf)
{
   if (avr_init ())
   {
      avr_run ();
      return -1;
   }

   /* Low fuse-byte */
   sendByte (0x50);
   sendByte (0x00);
   sendByte (0x00);
   buf[0] = (unsigned char) sendByte (0x00);

   /* High fuse-byte */
   sendByte (0x58);
   sendByte (0x08);
   sendByte (0x00);
   buf[1] = (unsigned char) sendByte (0x00);

   /* Also, read lock-byte */
   sendByte (0x58);
   sendByte (0x00);
   sendByte (0x00);
   buf[2] = ((unsigned char) sendByte (0x00)) & 0x3F;

   avr_run ();

   return 0;
}

/** avr_writefuses( unsigned char *buf)
 * 
 *  Write fuse bytes to AVR
 */
int
avr_writefuses (unsigned char *buf)
{
   if (avr_init ())
   {
      avr_run ();
      return -1;
   }

   /* Low fuse-byte */
   sendByte (0xAC);
   sendByte (0xA0);
   sendByte (0x00);
   sendByte (buf[0]);
   waitms (20);

   /* High fuse-byte */
   sendByte (0xAC);
   sendByte (0xA8);
   sendByte (0x00);
   sendByte (buf[1]);
   waitms (20);

   avr_run ();

   return 0;
}

/*
 */
int
avr_read (char *buf, unsigned int address, unsigned int len, int memory)
{
   int i;

   for (i = 0; i < (int) len; i++)
   {
      sendByte (memory << 7 |
                (((!FLAG_ENDIAN & !(i & 1)) |
                  (FLAG_ENDIAN & (i & 1))) & !memory) << 3 | 0x20);
      sendByte ((address + i) / (2 - memory) / 256);
      sendByte ((address + i) / (2 - memory) % 256);
      buf[i] = (unsigned char) sendByte (0x00);
   }

   return 0;
}

int
download (int nrOfBytes, unsigned char *data, int memory, unsigned int address)
{
   int i;
   int debug;
   unsigned char readByte;
   unsigned char P1, P2;
   int succes;

   if (memory == FLASH)
   {
      P1 = this_device->P3;
      P2 = P1;
   }
   else
   {
      P1 = this_device->P1;
      P2 = this_device->P2;
   }

   fprintf (stdout, "Programming %d bytes at %2.2X/%2.2X... ", nrOfBytes,
            address / 256, address % 256);
   for (i = 0; i < nrOfBytes; i++)
   {
      /* memory decides on FLASH / EEPROM if ENDIAN is selected and FLASH is 
       * programmed high and low order bytes are reversed -> odd bytes are
       * programmed at high byte. */
      /*
       * sendByte (memory << 7 |
       * (((!FLAG_ENDIAN & !(i & 1)) |
       * (FLAG_ENDIAN & (i & 1))) & !memory) << 3 | 0x40);
       *
       * Adjustment because MEGA-devices require first low then high byte: */
      if (!FLAG_ENDIAN)
      {
         if (i % 2 == 0)
         {
            i++;
         }
         else
         {
            i--;
         }
      }
      sendByte (memory << 7 |
                (((!FLAG_ENDIAN & !(i & 1)) |
                  (FLAG_ENDIAN & (i & 1))) & !memory) << 3 | 0x40);
      /* The address to program to: FLASH is 16-bit address, EEPROM is 8-bit
       * address. Each FLASH-instruction has 2 bytes at 1 address each EEPROM 
       * byte has only 1 byte per address. */
      sendByte ((address + i) / (2 - memory) / 256);
      sendByte ((address + i) / (2 - memory) % 256);
      /* Send the actual data */
      sendByte (data[i]);

      /* Difference between architectures: some (older) devices are
       * programmed per word, the MEGAs are programmed per page, thus only
       * after last byte */
      if (!this_device->pageprog || memory == EEPROM)
      {
         /* Poll the memory location until it is equal to the data
          * programmed to it, or wait 8 ms if it is the same as a control
          * character. */
         debug = 0;
         if (data[i] != P1 && data[i] != P2)
         {
            do
            {
               sendByte (memory << 7 |
                         (((!FLAG_ENDIAN & !(i & 1))
                           | (FLAG_ENDIAN & (i & 1))) & !memory) << 3 | 0x20);
               sendByte ((address + i) / (2 - memory) / 256);
               sendByte ((address + i) / (2 - memory) % 256);
               readByte = sendByte (0x00);
               debug++;
            }
            while (debug < 10000 && (readByte == P1 || readByte == P2));
         }
         else
         {
            /* The specified 8 ms seems not always long enough */
            waitms (20);
         }
         if (debug == 10000)
            fprintf (stderr,
                     "Error occurred with char at %02X/%02X:%02X\n",
                     address / 256, address % 256, i);
      }
      if (!FLAG_ENDIAN)
      {
         if (i % 2 == 0)
         {
            i++;
         }
         else
         {
            i--;
         }
      }
   }                            /* for-loop */
   /* In case of MEGAs do page-write */
   if (this_device->pageprog && memory == FLASH)
   {
      sendByte (0x4C);
      sendByte ((address) / (2 - memory) / 256);
      sendByte ((address) / (2 - memory) % 256);
      sendByte (0x00);

      /* Do not poll: too complicated */
      waitms (20);
   }

   /* Verify */
   if (FLAG_NOVERIFY)
   {
      fprintf (stdout, "(not verified)\x0D");
      fflush (stdout);
      return 0;
   }

   fprintf (stdout, "Verifying... ");
   succes = 1;
   for (i = 0; i < nrOfBytes; i++)
   {
      unsigned char temp;

      sendByte (memory << 7 |
                (((!FLAG_ENDIAN & !(i & 1)) |
                  (FLAG_ENDIAN & (i & 1))) & !memory) << 3 | 0x20);
      sendByte ((address + i) / (2 - memory) / 256);
      sendByte ((address + i) / (2 - memory) % 256);
      if ((temp = (unsigned char) sendByte (0x00)) != data[i])
      {
         printf ("(%X) Was: %X, should have been %X\n", i, temp, data[i]);
         succes = 0;
      }
   }
   if (!succes)
   {
      fprintf (stdout, "ERROR!\n");
      return -1;
   }

   fprintf (stdout, "OK!\x0D");
   fflush (stdout);
   return 0;
}

/*
 * int convert( int memory, char *s)
 * 
 * This function converts the given IHEX-formatted string to data.
 * s contains the Intel-HEX encoded string to download.
 * Returns: 
 * #bytes   on succes, 
 * -1       on failure. 
 */
int
convert (int memory, char *s, unsigned char *data, int *address)
{

   unsigned char nrOfBytes, checksum, type;
   unsigned char check;
   int i;

   if (s[0] != ':')
      not_OK ("Invalid file-format!\n")
         /* The number of bytes to program */
         nrOfBytes = hex2dec (s[1], s[2]);
   /* Start address, beware: This is the start-address in BYTES, not WORDS */
   *address = hex2dec (s[3], s[4]) * 256 + hex2dec (s[5], s[6]);
   /* Type of field, 0 means data-field */
   type = hex2dec (s[7], s[8]);
   /* Calculate checksum up to this point */
   check = nrOfBytes + (char) (*address / 256) + (char) (*address % 256) + type;

   /* Check address range */
   if ((memory == FLASH && *address > this_device->FLASHsize) ||
       (memory == EEPROM && *address > this_device->EEPROMsize))
   {
      not_OK ("Address range exceeds space of device!\n");
   }

   if (type == 0)
   {
      for (i = 0; i < nrOfBytes; i++)
      {
         data[i] = hex2dec (s[9 + i * 2], s[10 + i * 2]);
         check += data[i];
      }
   }
   else
   {
      /* Only data fields are programmed. */
      return 0;
   }

   /* Read the real checksum from the file */
   checksum = hex2dec (s[9 + nrOfBytes * 2], s[10 + nrOfBytes * 2]);

   /* Compare the checksum in the file with the calculated checksum */
   if ((unsigned char) checksum != (unsigned char) (256 - check))
   {
      //      not_OK ("Checksum ERROR!\n");
   }

   /* If everything OK, then return number of bytes */
   return nrOfBytes;
}

/** int avr_init( unsigned int *ID)
 * 
 * Initializes the AVR for programming.
 * Returns: 
 * 0  on succes
 * -1 on failure to talk to AVR
 */
int
avr_init (void)
{
   int i;

   /* Clear the parallel port. This also enters RESET mode of the AVR. */
   outb (0x00, LPT_BASE);

   /* Send programming Enable, at most 32 times, if not succesfull after 32
    * tries, then quit. */
   for (i = 0; i < 32; i++)
   {
      sendByte (0xAC);
      sendByte (0x53);
      if ((sendByte (0x00) == 0x53))
      {
         sendByte (0x00);
         break;
      }
      sendByte (0x00);
      /* Short _reset_ pulse and try again */
      outb (1 << RESET, LPT_BASE);
      waitms (20);
      outb (0x00, LPT_BASE);
   }
   if (i == 32)
   {
      printf ("Init failed!!!!\n");
      return -1;
   }
   return 0;
}

/** avr_run()
 *  
 * Remove RESET signal
 */
int
avr_run (void)
{
   /* Turn _RESET off */
   if (ioperm (LPT_BASE, 3, 0x666) != -1)
   {
      outb ((1 << RESET) | (1 << SCK) | (1 << MOSI), LPT_BASE);
   }
   return 0;
}

/** avr_readID( *device)
 *
 *  Return the currently found device;
 *  Return 0  - succes
 *         -1 - com-failure
 *         -2 - ID failure
 */
int
avr_readID (struct ID_CODES **device)
{
   int i;
   unsigned int ID[3];

   if (avr_init ())
   {
      avr_run();
      return -1;
   }

   /* Ask for the three identification bits */
   for (i = 0; i < 3; i++)
   {
      sendByte (0x30);
      sendByte (0x00);
      sendByte (0x00 + i);
      ID[i] = (unsigned char) sendByte (0x00);
   }

   /* Find these bits in the list and determine type */
   *device = NULL;
   for (i = 0; id_codes[i].ID; i++)
   {
      if (id_codes[i].ID == (ID[0] << 16 | ID[1] << 8 | ID[2]))
      {
         *device = &(id_codes[i]);
         break;
      }
   }

   avr_run ();

   if (*device == NULL)
   {
      fprintf( stderr, "Couldn't identify device, ID=%02X:%02X:%02X\n", ID[0], ID[1], ID[2]);
      return -2;
   }
   return 0;
}

/*
 * unsigned char sendByte( unsigned char b)
 * 
 * Sends the byte b to the AVR and reads the output of the AVR at the same
 * clock. Used to send bytes, to read bytes _and_ for the init-procedure 
 * send/receive
 */
unsigned char
sendByte (unsigned char b)
{
   char res = 0;
   int i;
   int temp;

   for (i = 7; i >= 0; i--)
   {
      /* SCK should currently be high, a high->low transition will read a
       * character from MISO, value of MOSI can be anything (0). */
      outb (0x00, LPT_BASE);
      temp = inb (LPT_BASE + 1);
#ifndef REVMISO
      res = res << 1 | ((temp & (1 << MISO)) >> MISO);
#else
      res = res << 1 | !((temp & (1 << MISO)) >> MISO);
#endif

      /* Set up value on MOSI */
      outb ((((b & (1 << i)) >> i) << MOSI), LPT_BASE);

      /* Data bit and low->high clock, this will clock the data in */
      outb ((((b & (1 << i)) >> i) << MOSI) | (1 << SCK), LPT_BASE);
      
      /* Apparently, some delay is needed here (10ns) */
      waitus( 1);
   }

   /* Note: SCK is still high, this will enable a READ next time this
    * function is called. */

   /* Return the byte read from the input pin: */
   return res;
}

/* int avr_erase( void)
 * 
 * Erase contents of both flash and EEPROM of the AVR.
 * Returns:
 * 0  on succes.
 * -1 on failaure (not able to synchronize again).
 */
int
avr_erase (void)
{
   int i;

   /* Chip Erase */
   sendByte (0xAC);
   sendByte (0x80);
   sendByte (0x00);
   sendByte (0x00);

   /* Wait 18 ms */
   waitms (18);

   /* Short _reset_ pulse */
   outb (0x04, LPT_BASE);
   waitms (10);
   outb (0x00, LPT_BASE);

   /* Wait 20 ms */
   waitms (10);

   /* Send programming Enable (again), at most 32 times, if not succesfull
    * after 32 tries, then quit. */
   for (i = 0; i < 32; i++)
   {
      sendByte (0xAC);
      sendByte (0x53);
      if (sendByte (0x00) == 0x53)
      {
         sendByte (0x00);
         break;
      }
      sendByte (0x00);
      /* Short _reset_ pulse and try again */
      outb (0x04, LPT_BASE);
      waitms (10);
      outb (0x00, LPT_BASE);
   }
   if (i == 32)
      return -1;

   return 0;
}

/*
 * Waits for approx. w ms (in multiples of 1/TCK_CLK)
 * TCK_CLK is in my case 100, so in multiples of 10 ms.
 */
void
waitms (clock_t w)
{
   clock_t start;
   
   w = (clock_t) ceil ((double) w / 10.0) * 10;
   start = times (NULL);
   while ((times (NULL) - start) * 1000 < (w * TCK_CLK))
      ;
}

/*
 * Waits for at least w microseconds
 */
void
waitus( long w)
{
   static struct timeval tv0, tv;
      
      gettimeofday( &tv0, NULL);
      tv = tv0;
      while( (tv.tv_sec-tv0.tv_sec)*1000000 + (tv.tv_usec - tv0.tv_usec) < w)
      {
         gettimeofday( &tv, NULL);
   }
}

/*
 * void terminate( int exitcode, char *exitmsg)
 * 
 * Displays the message exitmsg and quits with code exitcode.
 * Also removes any RESET on AVR from parallel port.
 */
void
terminate (int exitcode, char *exitmsg)
{
   /* Turn _RESET off */
   if (ioperm (LPT_BASE, 3, 0x666) != -1)
   {
      outb ((1 << RESET) | (1 << SCK) | (1 << MOSI), LPT_BASE);
   }

   /* display message and quit */
   fprintf (stderr, "%s\n", exitmsg);
   exit (exitcode);
}
